package com.supervision.police.service.impl;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.supervision.chat.UploadParamEnum;
import com.supervision.chat.client.CustomMultipartFile;
import com.supervision.chat.client.LangChainChatService;
import com.supervision.chat.client.dto.DeleteFileDTO;
import com.supervision.chat.client.dto.LangChainChatRes;
import com.supervision.common.utils.IPages;
import com.supervision.common.utils.ListUtils;
import com.supervision.common.utils.StringUtils;
import com.supervision.config.BusinessException;
import com.supervision.demo.dto.QARecordNodeDTO;
import com.supervision.minio.domain.MinioFile;
import com.supervision.minio.mapper.MinioFileMapper;
import com.supervision.minio.service.MinioService;
import com.supervision.neo4j.service.Neo4jService;
import com.supervision.police.domain.*;
import com.supervision.police.dto.NoteRecordDTO;
import com.supervision.police.dto.NoteRecordDetailDTO;
import com.supervision.police.mapper.NoteRecordMapper;
import com.supervision.police.mapper.NoteRecordSplitMapper;
import com.supervision.police.service.*;
import com.supervision.utils.RecordRegexUtil;
import com.supervision.utils.WordReadUtil;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.io.IOException;
import java.io.InputStream;
import java.time.LocalDateTime;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

@Slf4j
@Service
@RequiredArgsConstructor
public class NoteRecordSplitServiceImpl extends ServiceImpl<NoteRecordSplitMapper, NoteRecordSplit> implements NoteRecordSplitService {

    private final NoteRecordMapper noteRecordMapper;

    private final MinioService minioService;

    private final MinioFileMapper minioFileMapper;

    @Autowired
    private NoteRecordSplitService noteRecordSplitService;


    private final CaseTaskRecordService caseTaskRecordService;

    private final TripleInfoService tripleInfoService;

    @Autowired
    private ModelRecordTypeService modelRecordTypeService;

    @Autowired
    private RecordSplitProcessService recordSplitProcessService;

    private final CaseStatusManageService caseStatusManageService;

    private final Neo4jService neo4jService;

    private final LangChainChatService langChainChatService;


    @Override
    @Transactional(transactionManager = "dataSourceTransactionManager", rollbackFor = Exception.class)
    public String uploadRecords(NoteRecordDTO records) throws IOException {
        //修改案件的行为人、当事人
        /*if (StringUtils.isNotEmpty(records.getRole()) && !"嫌疑人".equals(records.getRole())) {
            ModelCase modelCase = modelCaseMapper.selectById(records.getCaseId());
            //行为人
            if ("行为人".equals(records.getRole()) && StringUtils.isEmpty(modelCase.getLawActor())) {
                modelCase.setLawActor(records.getName());
                modelCaseMapper.updateById(modelCase);
            } else if ("受害人".equals(records.getRole()) || "证人".equals(records.getRole())) {
                if (StringUtils.isEmpty(modelCase.getLawParty())) {
                    modelCase.setLawParty(records.getName());
                } else {
                    String[] split = modelCase.getLawParty().split(",");
                    if (!Arrays.asList(split).contains(records.getName())) {
                        modelCase.setLawParty(modelCase.getLawParty() + "," + records.getName());
                    }
                }
                modelCaseMapper.updateById(modelCase);
            }
        }*/
        int i;
        NoteRecord record = records.toNoteRecord();
        if (StringUtils.isEmpty(records.getId())) {
            i = noteRecordMapper.insert(record);
        } else {
            i = noteRecordMapper.updateById(record);
        }
        //所有对话类型
        List<ModelRecordType> allTypeList = modelRecordTypeService.lambdaQuery().list();
        if (i > 0 && CollUtil.isNotEmpty(records.getFileIdList())) {
            // 拆分笔录
            for (String fileId : records.getFileIdList()) {
                MinioFile minioFile = minioService.getMinioFile(fileId);
                InputStream inputStream = null;
                try {
                    inputStream = minioService.getObjectInputStream(minioFile);
                } catch (Exception e) {
                    e.printStackTrace();
                    log.error("从minio中获取文件失败:{}", e.getMessage());
                    continue;
                }
                // 这里需要把文件传输,传输到知识库中去
                CustomMultipartFile mockMultipartFile = new CustomMultipartFile(minioFile.getFilename(), inputStream);
                LangChainChatRes langChainChatRes = langChainChatService.uploadFile("11111111",
                        mockMultipartFile,
                        "问讯笔录",
                        UploadParamEnum.to_vector_store.getBooleanValue(),
                        UploadParamEnum.override.getBooleanValue(),
                        UploadParamEnum.not_refresh_vs_cache.getBooleanValue(),
                        UploadParamEnum.chunk_size.getIntValue(),
                        UploadParamEnum.chunk_overlap.getIntValue(),
                        UploadParamEnum.zh_title_enhance.getBooleanValue(),
                        UploadParamEnum.docs.getStrValue());
                if (!langChainChatRes.getCode().equals(200)) {
                    log.error("上传文件失败:{}", langChainChatRes.getMsg());
                    throw new RuntimeException("上传文件失败");
                }
                String context = WordReadUtil.readWord(inputStream);
                List<QARecordNodeDTO> qaList = RecordRegexUtil.recordRegex(context, record.getName());
                List<NoteRecordSplit> splitList = new ArrayList<>();
                for (QARecordNodeDTO qa : qaList) {
                    try {
                        NoteRecordSplit noteRecord = new NoteRecordSplit();
                        noteRecord.setCaseId(record.getCaseId());
                        noteRecord.setNoteRecordId(record.getId());
                        noteRecord.setNoteName(minioFile.getFilename());
                        noteRecord.setPersonName(record.getName());
                        noteRecord.setQuestion(qa.getQuestion());
                        noteRecord.setAnswer(qa.getAnswer());
                        noteRecord.setCreateTime(LocalDateTime.now());
                        this.save(noteRecord);
                        splitList.add(noteRecord);

                    } catch (Exception e) {
                        log.error(e.getMessage(), e);
                    }
                }
                // 创建任务之后,再开始
                boolean taskStatus = saveRecordProcessTask(record.getCaseId(), record.getId());
                // 如果校验结果为false,则说明需要进行分类以及三元组操作
                if (taskStatus) {
                    // 对笔录进行分类,并对笔录进行提取三元组
                    recordSplitProcessService.process(allTypeList, splitList);
                }
            }
            caseStatusManageService.whenUploadRecord(record.getCaseId());
            return "保存成功";
        } else {
            return "保存笔录失败";
        }
    }

    /**
     * 校验是否已经执行过任务
     *
     * @param caseId
     * @param recordId
     * @return 没有执行过, 返回true, 没有执行过, 返回false
     */
    private boolean saveRecordProcessTask(String caseId, String recordId) {
        // 首先查询是否存在任务,如果不存在,就新建
        Optional<CaseTaskRecord> caseTaskRecordOpt = caseTaskRecordService.lambdaQuery()
                .eq(CaseTaskRecord::getCaseId, caseId).eq(CaseTaskRecord::getRecordId, recordId).oneOpt();
        if (caseTaskRecordOpt.isEmpty()) {
            CaseTaskRecord newCaseTaskRecord = new CaseTaskRecord();
            newCaseTaskRecord.setCaseId(caseId);
            newCaseTaskRecord.setRecordId(recordId);
            newCaseTaskRecord.setStatus(1);
            newCaseTaskRecord.setSubmitTime(LocalDateTime.now());
            return caseTaskRecordService.save(newCaseTaskRecord);
        } else {
            // 如果存在,则校验时间是否已经超过1天,如果超过了1天还没有执行完毕,就重新提交这个任务
            CaseTaskRecord caseTaskRecord = caseTaskRecordOpt.get();
            // 如果未执行,则提交执行
            if (caseTaskRecordOpt.get().getStatus() == 0) {
                caseTaskRecord.setStatus(1);
                caseTaskRecord.setSubmitTime(LocalDateTime.now());
                caseTaskRecordService.updateById(caseTaskRecord);
                return true;
            }
            if (caseTaskRecordOpt.get().getStatus() == 1 && LocalDateTime.now().isAfter(caseTaskRecord.getSubmitTime().plusDays(1))) {
                // 如果已经超过1天,则重新提交任务
                caseTaskRecord.setStatus(1);
                caseTaskRecord.setSubmitTime(LocalDateTime.now());
                caseTaskRecordService.updateById(caseTaskRecord);
                return true;
            } else if (caseTaskRecordOpt.get().getStatus() == 2) {
                // 如果执行成功,就不再执行了
                return false;
            } else if (caseTaskRecordOpt.get().getStatus() == 3) {
                // 如果执行失败,就重新执行
                caseTaskRecord.setStatus(1);
                caseTaskRecord.setSubmitTime(LocalDateTime.now());
                caseTaskRecordService.updateById(caseTaskRecord);
                return true;
            } else {
                // 如果都不是就在跑一次
                caseTaskRecord.setStatus(1);
                caseTaskRecord.setSubmitTime(LocalDateTime.now());
                caseTaskRecordService.updateById(caseTaskRecord);
                return true;

            }
        }
    }


    @Override
    public Map<String, Object> queryRecords(NoteRecord noteRecords, Integer page, Integer size) {
        LambdaQueryWrapper<NoteRecord> wrapper = Wrappers.lambdaQuery(NoteRecord.class)
                .like(StringUtils.isNotEmpty(noteRecords.getName()), NoteRecord::getName, noteRecords.getName())
                .eq(NoteRecord::getCaseId, noteRecords.getCaseId())
                .eq(NoteRecord::getDataStatus, "1");
        List<NoteRecord> list = noteRecordMapper.selectList(wrapper);
        List<String> fileIdList = list.stream().map(NoteRecord::getFileIds).filter(StringUtils::isNotEmpty)
                .flatMap(s -> Arrays.stream(s.split(","))).distinct().toList();

        Map<String, MinioFile> fileMap = minioService.listMinioFile(fileIdList).stream().collect(Collectors.toMap(MinioFile::getId, Function.identity()));

        // LinkedHashMap 保障顺序
        Map<String, List<NoteRecordDetailDTO>> nameMap = list.stream().filter(item -> StringUtils.isNotBlank(item.getName())).map(item -> new NoteRecordDetailDTO(item, fileMap))
                .collect(Collectors.groupingBy(NoteRecordDetailDTO::getName, LinkedHashMap::new, Collectors.toList()));
        // 获取所有的task任务
        List<CaseTaskRecord> taskList = caseTaskRecordService.lambdaQuery().eq(CaseTaskRecord::getCaseId, noteRecords.getCaseId()).list();
        Map<String, CaseTaskRecord> taskRecordMap = taskList.stream().collect(Collectors.toMap(CaseTaskRecord::getRecordId, Function.identity(), (k1, k2) -> k1));

        List<NoteRecordDetailDTO> res = new ArrayList<>();
        for (String name : nameMap.keySet()) {
            NoteRecordDetailDTO noteRecordDetailDTO = new NoteRecordDetailDTO();
            noteRecordDetailDTO.setName(name);

            List<NoteRecordDetailDTO> noteRecordDetailDTOS = nameMap.get(name);
            int fileCount = 0;
            for (NoteRecordDetailDTO recordDetailDTO : noteRecordDetailDTOS) {
                fileCount = fileCount + recordDetailDTO.getFileList().size();
                CaseTaskRecord caseTaskRecord = taskRecordMap.get(recordDetailDTO.getId());
                if (ObjectUtil.isNotEmpty(caseTaskRecord)) {
                    if (caseTaskRecord.getStatus() == 2) {
                        recordDetailDTO.setPercentage("100%");
                    } else if (caseTaskRecord.getStatus() == 0) {
                        recordDetailDTO.setPercentage("0%");
                    } else {
                        if (caseTaskRecord.getTaskCount() == 0) {
                            recordDetailDTO.setPercentage("0%");
                        } else {
                            // 获取执行的数量
                            double v = caseTaskRecord.getFinishCount() * 100.0 / caseTaskRecord.getTaskCount();
                            // 不能超过100%
                            recordDetailDTO.setPercentage(String.format("%.2f%%", Math.min(v, 100)));
                        }
                    }
                }
            }
            noteRecordDetailDTO.setChildren(noteRecordDetailDTOS);
            // 统计所有children的文件数量
            noteRecordDetailDTO.setConfessionMaterial(fileCount);
            res.add(noteRecordDetailDTO);
        }
        List<NoteRecordDetailDTO> pager = ListUtils.Pager(size, page, res);
        return IPages.buildDataMap(pager, res.size());
    }

    @Override
    @Transactional(transactionManager = "dataSourceTransactionManager", rollbackFor = Exception.class, noRollbackFor = BusinessException.class)
    public void delRecords(String id) {

        NoteRecord noteRecord = noteRecordMapper.selectById(id);
        // 去掉逻辑删除,直接删除
        // noteRecord.setDataStatus(StringUtils.getUUID());
        // noteRecordMapper.updateById(noteRecord);
        String fileIds = noteRecord.getFileIds();
        if (StringUtils.isNotEmpty(fileIds)) {

            //删除文件
            for (String fileId : fileIds.split(",")) {

                MinioFile minioFile = minioFileMapper.selectById(fileId);
                // 删除文件
                LangChainChatRes langChainChatRes = langChainChatService.deleteFile(DeleteFileDTO.create(noteRecord.getCaseId(), minioFile.getFilename()));
                if (!langChainChatRes.getCode().equals(200)) {
                    throw new BusinessException("删除文件:" + minioFile.getFilename() + " 失败!");
                }
                minioFile.setDataStatus(StringUtils.getUUID());
                minioFileMapper.updateById(minioFile);
                minioService.delFile(fileId);

            }
        }
        // 去掉逻辑删除,直接删除
        noteRecordMapper.deleteById(id);
        // 删除拆分的笔录
        noteRecordSplitService.lambdaUpdate().eq(NoteRecordSplit::getNoteRecordId, id).remove();
        // task也删除掉
        caseTaskRecordService.lambdaUpdate().eq(CaseTaskRecord::getRecordId, id).remove();
        // 更新笔录的状态
        caseStatusManageService.whenDeleteRecord(noteRecord.getCaseId());
        // 删除neo4j的节点
        List<TripleInfo> list = tripleInfoService.lambdaQuery().eq(TripleInfo::getRecordId, id).list();
        list.forEach(item -> {
            // 如果已经入库,就删除已经入库的图
            if (StrUtil.equals("1", item.getAddNeo4j())) {
                // 删除尾节点(不删除头节点),如果删除尾节点,关系会被自动删除
                try {
                    neo4jService.delNode(item.getEndNodeGraphId());
                } catch (Exception e) {
                    log.error("删除关系失败:{}", item.getRelation(), e);
                }

            }

        });
        // TODO 是否需要把model_atomic_result的结果也删除了?
    }

}
